"""Top level project/package init."""
from importlib import metadata


def version():
    """Return the canonical version from pyproject.toml used to package this particular release."""
    return metadata.version("raindrop_io_py")



"""Display help."""
from rich.table import Table

from raindropiopy.cli import COLOR_TABLE_COLUMN_1, COLOR_TABLE_COLUMN_2
from raindropiopy.cli.models.eventLoop import EventLoop


def process(el: EventLoop) -> None:
    """Display help."""
    el.console.print("Help is here, never fear!")


def help_search(el: EventLoop) -> None:
    """Render and display search help."""
    contents = (
        (
            "*",
            "All Raindrops (but you must specify at least one collection)",
        ),
        (
            "apple iphone",
            "Find items that contains such words in title, description, domain or in web page content",
        ),
        ("#coffee", "Find items that have a certain tag"),
        (
            "'superman vs. batman'",
            "Find items that contains exact phrase in title, description, domain or in web page content",
        ),
        (
            "-superman -#coffee",
            "Requires that the search results do not include this word or condition, eg. no superman, not tag coffee.",
        ),
        ("superman batman match:OR", "Find items with either search term"),
        ("created:2021-07-15 ", "Search for items created on a specific date"),
        ("created:2021-07 ", "Search for items created in a specific month"),
        ("created:2021 ", "Search for items created in a specific year"),
        ("created:>2021-07-15", "Put > in front of a date to find after specific date"),
        (
            "created:<2021-07-15",
            "Put < in front of a date to find before specific date",
        ),
        ("lastUpdate:2021-07-15", "Search for items updated in specific date"),
        ("link:dropbox", "Find items with a certain word (or words) in the URL"),
        ("link:'crunch base'", "Find items with a certain word (or words) in the URL"),
        ("type:link", "Find by type"),
        ("type:article", "Find by type"),
        ("type:image", "Find by type"),
        ("type:video", "Find by type"),
        ("type:document", "Find by type"),
        ("type:audio", "Find by type"),
        ("file:true", "Find files"),
        ("notag:true", "Find items without tags"),
    )
    table = Table(title=None, show_header=False)
    table.add_column("Search", style=COLOR_TABLE_COLUMN_1)
    table.add_column("Descript", style=COLOR_TABLE_COLUMN_2)
    for tpl in contents:
        table.add_row(*tpl)
    el.console.print(table)



"""Create a new Raindrop bookmark."""
import webbrowser
from typing import Final

from prompt_toolkit.completion import WordCompleter
from rich.table import Table

from raindropiopy.cli import (
    COLOR_TABLE_COLUMN_1,
    COLOR_TABLE_COLUMN_2,
    PROMPT_STYLE,
    PROMPT_STYLE_WARNING,
    WARNING,
    prompt,
    options_as_help,
)
from raindropiopy.api.models import Raindrop
from raindropiopy.cli.commands import get_confirmation, get_from_list, is_int
from raindropiopy.cli.models.eventLoop import EventLoop
from raindropiopy.cli.models.searchState import SearchState
from raindropiopy.cli.models.spinner import Spinner
from raindropiopy.cli.commands import get_title, get_description


def _prompt_view_edit(el: EventLoop, search_state: SearchState) -> str:
    """Prompt for the next thing to do with the specified raindrop."""
    while True:
        options: Final = ("open", "edit", "view", "list", "delete", "requery", "back/.")
        options_title: Final = options_as_help(options)
        el.console.print(options_title)
        response = el.session.prompt(
            prompt(("search results", search_state.prompt())),
            completer=WordCompleter(options),
            style=PROMPT_STYLE,
            complete_while_typing=True,
            enable_history_search=False,
        )
        if is_int(response) and search_state.results:
            i_response = int(response)
            if 1 <= i_response <= len(search_state):
                return int(response)
            else:
                print(
                    f"[{WARNING}]Sorry, raindrop selector must be between 1 and {len(search_state)}.",
                )
                continue
        return response


def _view_raindrop(el: EventLoop, raindrop: Raindrop) -> None:
    """Display the currently selected raindrop."""
    table = Table(show_header=False)
    table.add_column(
        "Attribute",
        justify="left",
        style=COLOR_TABLE_COLUMN_1,
        no_wrap=True,
    )
    table.add_column("Value", justify="left", style=COLOR_TABLE_COLUMN_2)

    # Disambiguate some attributes for easier viewing:
    raindrop._collection = el.state.find_collection_by_id(raindrop.collection.id).title
    raindrop._tags = ", ".join(raindrop.tags)
    raindrop._type = raindrop.type.value.title()
    raindrop._created = raindrop.created.isoformat().split("T")[0]
    raindrop._updated = raindrop.lastUpdate.isoformat().split("T")[0]
    if raindrop._updated == raindrop._created:
        raindrop._updated = None

    for attr, title in [
        ["_collection", "Collection"],
        ["title", "Title"],
        ["excerpt", "Description"],
        ["_tags", "Tags"],
        ["_type", "Type"],
        ["_created", "Created"],
        ["_updated", "Updated"],
    ]:
        if value := getattr(raindrop, attr, "-"):
            table.add_row(title, str(value))

    el.console.print(table)


def _edit_raindrop(
    el: EventLoop,
    search_state: SearchState,
    raindrop: Raindrop,
) -> None:
    """Edit selected attributes of the currently selected raindrop."""
    _view_raindrop(el, raindrop)

    # What attribute to edit?
    search_context = search_state.prompt()
    options: Final = ("title", "tags", "description", "back/.")
    options_title: Final = options_as_help(options, 2)
    el.console.print(options_title)
    prompts = ("search results", search_context, "edit")
    response = el.session.prompt(
        prompt(prompts),
        completer=WordCompleter(options),
        style=PROMPT_STYLE,
        complete_while_typing=True,
        enable_history_search=False,
    )
    if response.casefold() in ("b", "back", "."):
        return

    elif response.casefold().startswith("ta"):
        prompts = ("search results", search_context, "edit", "tag(s)?")
        tags = get_from_list(el, prompts, list(el.state.tags))
        with Spinner("Updating Raindrop..."):
            Raindrop.update(el.state.api, id=raindrop.id, tags=tags)

    elif response.casefold().startswith("ti"):
        prompts = ("search results", search_context, "edit", "title?")
        title = get_title(el, prompts)
        with Spinner("Updating Raindrop..."):
            Raindrop.update(el.state.api, id=raindrop.id, title=title)

    elif response.casefold().startswith("d"):
        prompts = ("search results", search_context, "edit", "description?")
        excerpt = get_description(el, prompts)
        with Spinner("Updating Raindrop..."):
            Raindrop.update(el.state.api, id=raindrop.id, excerpt=excerpt)

    return


def _open_raindrop(el: EventLoop, raindrop: Raindrop) -> None:
    """Open the currently selected raindrop."""
    webbrowser.open(raindrop.link)


def _delete_raindrop(el: EventLoop, raindrop: Raindrop) -> None:
    """Delete the currently selected raindrop (after confirmation)."""
    _view_raindrop(el, raindrop)
    if get_confirmation(el, "Are you sure?", PROMPT_STYLE_WARNING):
        with Spinner(f"Deleting Raindrop: [italic]{raindrop.title}[/]..."):
            Raindrop.remove(el.state.api, id=raindrop.id)


def process(el: EventLoop, search_state: SearchState) -> bool:
    """Top-level UI Controller for viewing / edit a bookmark(s).

    Return:
    - quit : Should we "quit" after this invocation?
    """
    while True:
        response = _prompt_view_edit(el, search_state)

        if is_int(response):
            raindrop = search_state.get_selected(int(response))
            _view_raindrop(el, raindrop)

        elif response.casefold() in ("b", "back", ".", "q", "quit"):
            return True

        elif response.casefold() in ("v", "view"):
            _view_raindrop(el, raindrop)

        elif response.casefold() in ("o", "open"):
            _open_raindrop(el, raindrop)

        elif response.casefold() in ("l", "list"):
            search_state.display_results(el)

        elif response.casefold() in ("r", "requery"):
            search_state.query(el)
            search_state.display_results(el)

        elif response.casefold() in ("d", "delete"):
            _delete_raindrop(el, raindrop)
            search_state.query(el)
            search_state.display_results(el)

        elif response.casefold() in ("e", "edit"):
            _edit_raindrop(el, search_state, raindrop)
            search_state.query(el)
            search_state.display_results(el)



"""Low-level API interface to Raindrop, no application semantics, mostly core HTTP verbs."""
from __future__ import annotations

import datetime
import enum
import json
from pathlib import Path
from typing import Any, Dict, Optional, Union

import requests
from requests_oauthlib import OAuth2Session


class API:
    """Provides communication to the Raindrop.io API server.

    :param token: An access token for authorization.
    :type token: string or dict.
    """

    URL_AUTHORIZE = "https://raindrop.io/oauth/authorize"
    URL_ACCESS_TOKEN = "https://raindrop.io/oauth/access_token"
    URL_REFRESH = "https://raindrop.io/oauth/access_token"

    ratelimit: Optional[int] = None
    ratelimit_remaining: Optional[int] = None
    ratelimit_reset: Optional[int] = None

    def __init__(
        self,
        token: Union[str, Dict[str, Any]],
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token_type: str = "Bearer",
    ) -> None:
        """Instantiate an API connection to Raindrop using the credentials provided."""
        self.token = token
        self.client_id = client_id
        self.client_secret = client_secret
        self.token_type = token_type
        self.session = None

        self.open()

    def _create_session(self) -> OAuth2Session:
        extra: Optional[Dict[str, Any]]
        if self.client_id and self.client_secret:
            extra = {
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }
        else:
            extra = None

        def update_token(newtoken: str) -> None:
            self.token = newtoken

        if isinstance(self.token, str):
            token = {"access_token": self.token}
        else:
            token = self.token

        return OAuth2Session(
            self.client_id,
            token=token,
            auto_refresh_kwargs=extra,
            auto_refresh_url=self.URL_REFRESH,
            token_updater=update_token,
        )

    def open(self) -> None:
        """Open a new connection to Raindrop."""
        self.close()
        self.session = self._create_session()

    def close(self) -> None:
        """Close an existing Raindrop connection."""
        if self.session:
            self.session.close()
            self.session = None

    def _json_unknown(self, obj: Any) -> Any:
        if isinstance(obj, enum.Enum):
            return obj.value

        if isinstance(obj, datetime.datetime):
            return obj.isoformat()

        raise TypeError(
            f"Object of type {obj.__class__.__name__} is not JSON serializable",
        )

    def _to_json(self, obj: Any) -> Optional[str]:
        if obj is not None:
            return json.dumps(obj, default=self._json_unknown)
        else:
            return None

    def _on_resp(self, resp: Any) -> None:
        def get_int(name: str) -> Optional[int]:
            value = resp.headers.get(name, None)
            if value is not None:
                return int(value)
            return None

        v = get_int("X-RateLimit-Limit")
        if v is not None:
            self.ratelimit = v

        v = get_int("X-RateLimit-Remaining")
        if v is not None:
            self.ratelimit_remaining = v

        v = get_int("X-RateLimit-Reset")
        if v is not None:
            self.ratelimit_reset = v

        resp.raise_for_status()

    def _request_headers(self) -> Dict[str, str]:
        return {
            "Content-Type": "application/json",
        }

    def get(
        self,
        url: str,
        params: Optional[Dict[Any, Any]] = None,
    ) -> requests.models.Response:
        """Send a GET request.

        :param url: The url to send request
        :type url: str

        :param params: (optional) Dictionary, list of tuples or bytes to send
            in the query string for the :class:`Request`.

        :return: :class:`requests.Response` object
        :rtype: :class:`requests.Response`
        """
        assert self.session
        ret = self.session.get(url, headers=self._request_headers(), params=params)
        self._on_resp(ret)

        return ret

    def put(self, url: str, json: Any = None) -> requests.models.Response:
        """Low-level call to perform a PUT method against our present connection."""
        json = self._to_json(json)

        assert self.session
        ret = self.session.put(url, headers=self._request_headers(), data=json)
        self._on_resp(ret)
        return ret

    def put_file(
        self,
        url: str,
        path: Path,
        data: dict,
        files: dict,
    ) -> requests.models.Response:
        """Upload a file by a PUT request.

        :param url: The url to send request to
        :type url: str

        :param path: Path to file to be uploaded
        :type path: Path

        :param data: Dictionary, payload to be sent for the :class:`Request`,
            e.g. {"collectionId" : aCollection.id}
        :type data: dict

        :param files: Dictionary, request library "files" object to be sent for the :class:`Request`,
            e.g. {'file': (aFileName, aFileLikeObj, aContentType)}
        :type files: dict

        :return: :class:`requests.Response` object
        :rtype: :class:`requests.Response`
        """
        assert self.session
        ret = self.session.put(url, data=data, files=files)
        self._on_resp(ret)
        return ret

    def post(self, url: str, json: Any = None) -> requests.models.Response:
        """Low-level call to perform a POST method against our present connection."""
        json = self._to_json(json)

        assert self.session
        ret = self.session.post(url, headers=self._request_headers(), data=json)
        self._on_resp(ret)
        return ret

    def delete(self, url: str, json: Any = None) -> requests.models.Response:
        """Low-level call to perform a DELETE method against our present connection."""
        json = self._to_json(json)

        assert self.session
        ret = self.session.delete(url, headers=self._request_headers(), data=json)
        self._on_resp(ret)
        return ret

    def __enter__(self) -> API:
        """If we don't have an active session open yet, open one!."""
        if not self.session:
            self.open()
        return self

    def __exit__(self, _type, _value, _traceback) -> None:  # type: ignore
        """Once we're done with this API's scope, close the connection off."""
        self.close()



"""Top level cli.commands dunder init, mostly common methods and data types."""
from typing import Final, Optional

from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.styles import Style
from rich import print

from raindropiopy.cli import WARNING, PROMPT_STYLE, prompt
from raindropiopy.cli.models.eventLoop import EventLoop


def is_int(string: str) -> bool:
    """Make (compound) conditionals easier."""
    try:
        int(string)
        return True
    except ValueError:
        return False


def get_from_list(
    el: EventLoop,
    prompt_s: tuple[str],
    options: tuple[str],
) -> Optional[str]:
    """Mini-event loop to prompt for one or more selected options from the list provided."""
    completer: Final = WordCompleter(options)
    while True:
        response = el.session.prompt(
            prompt(prompt_s),  # ("create", "url", f"{prompt}?"))
            completer=completer,
            style=PROMPT_STYLE,
            complete_while_typing=True,
            enable_history_search=False,
        )
        if response == "?":
            el.console.print(", ".join(options))
        else:
            break
    return response


def get_title(el: EventLoop, prompts: list[str]) -> str | None:
    """Mini-event loop to prompt for a raindrop Title."""
    while True:
        title = el.session.prompt(prompt(prompts), style=PROMPT_STYLE)
        if title == "?":
            el.console.print(
                "We need a Raindrop title here, eg. 'Home Page for xxxyyy'",
            )
        elif title == "q":
            return None
        else:
            return title


def get_description(el: EventLoop, prompts: list[str]) -> str | None:
    """Mini-event loop to prompt for a raindrop Description (nee excerpt)."""
    while True:
        description = el.session.prompt(prompt(prompts), style=PROMPT_STYLE)
        if description == "?":
            el.console.print(
                "We need a Raindrop description here, eg. 'This was an interesting site.'",
            )
        elif description == "q":
            return None
        else:
            return description


def get_collection_s(el: EventLoop, prompts: list[str]) -> str | None:
    """Mini event-loop to prompt for one or many raindrop collections."""
    while True:
        collection_s = get_from_list(
            el,
            ("search", "collection(s)?"),
            el.state.get_collection_titles(exclude_unsorted=False),
        )
        if collection_s == ".":
            return None

        if collection_s:
            unrecognised_collections = list()
            current_collections = el.state.get_collection_titles(
                exclude_unsorted=True,
                casefold=True,
            )
            for collection in collection_s.split():
                if collection.casefold() not in current_collections:
                    unrecognised_collections.append(collection)
            if unrecognised_collections:
                print(
                    f"[{WARNING}]Sorry, collection(s) are NOT currently available: ",
                    end="",
                )
                print(", ".join(unrecognised_collections))
                print(
                    f"[{WARNING}]We're expecting either <blank> or one of one of: "
                    + ", ".join(el.state.get_collection_titles(exclude_unsorted=True)),
                )
                continue

        return collection_s


def get_confirmation(
    el: EventLoop,
    prompt: str,
    prompt_style: Style = PROMPT_STYLE,
) -> bool:
    """Ask the user for a confirmation to perform the specific prompt supplied."""
    el_prompt: Final = [("class:prompt", f"\n{prompt} ")]
    options: Final = ("yes", "Yes", "No", "no")
    completer: Final = WordCompleter(options)
    response = el.session.prompt(
        el_prompt,
        completer=completer,
        style=prompt_style,
        complete_while_typing=True,
        enable_history_search=False,
    )
    if response == "q":
        return None
    else:
        if response.lower() in ["yes", "y", "ye"]:
            return True
        return False



"""Spinner capability."""
#
# Inspired from beaupy (thanks Peter Výboch!) using underlying Rich-based "Live" class.
#
from itertools import cycle

from rich.live import Live

ARC = ["◜", "◠", "◝", "◞", "◡", "◟"]


class Spinner:
    """Rich-based CLI progress spinner as a Context Manager."""

    _spinner_characters: cycle
    _live_display: Live

    def __init__(self, text: str = "Loading...") -> None:
        """Create spinner for user feedback during a time-consuming operation.

        :param text: Text that will be shown after the spinner.

        Example:
        -------
            with Spinner("Please wait, I'm doing some work..."):
                do_something()
        """
        self._spinner_characters: list[str] = cycle(ARC)
        self._live_display = Live(
            "",
            transient=True,  # Remove spinner once complete.
            refresh_per_second=10,  # Reasonable default
            get_renderable=lambda: f"{next(self._spinner_characters)} {text}",
        )

    def __enter__(self):
        """Start the spinner."""
        self._live_display.start()

    def __exit__(self, _exc_type, _exc_value, _exc_tb):
        """Stop the spinner."""
        self._live_display.stop()



"""Startup/entry point for command-line interface."""
# This is meant to be packaged up as a module-level executable, thus, no hash-bang here.
# Still can be run directly though as: python raindropiopy/cli/cli.py
import os
import sys

from dotenv import load_dotenv

from raindropiopy.cli.models.eventLoop import EventLoop
from raindropiopy.cli.models.raindropState import RaindropState


def main():
    """Driver to kick off command-line interface."""
    load_dotenv()

    # Check that we have Raindrop connection token defined and available.
    if not os.getenv("RAINDROP_TOKEN"):
        sys.stderr.print(
            "Sorry, we need an environment variable (RAINDROP_TOKEN) to connect to the Raindrop.io service.\n",
        )
        sys.exit(1)

    EventLoop().go(RaindropState.factory())


if __name__ == "__main__":
    main()



"""Top level CLI dunder init, primarily constants used across all commands."""
import sys
from typing import Final

from prompt_toolkit.styles import Style

CONTENT_TYPES: Final = {
    ".pdf": "application/pdf",
    ".py": "text/plain",
    ".txt": "text/plain",
    ".org": "text/plain",
}


COLOR_TABLE_COLUMN_1 = "#00ffff"
COLOR_TABLE_COLUMN_2 = "#00ff00"
WARNING = "#fffc00"

PROMPT_STYLE: Final = Style.from_dict(
    {
        # We use our table color scheme to match
        "prompt": COLOR_TABLE_COLUMN_1,  # Prompt is cyan
        "": COLOR_TABLE_COLUMN_2,  # User input is green
    },
)

PROMPT_STYLE_WARNING: Final = Style.from_dict(
    {
        # We use our table color scheme to match
        "prompt": WARNING,  # Prompt is yellow
        "": WARNING,  # User input is yellow
    },
)


def prompt(sub_levels: tuple[str] | None = ()) -> str:
    """Render a command prompt to any number of 'levels'.

    With no sub-levels, prompt is: "> "
    With sub-levels ("A", "SubA"): "A> SubA> "
    """
    delimiter: str = "> "
    if sub_levels:
        prompt = ""
        for level in sub_levels:
            prompt += level + delimiter
    else:
        prompt = delimiter
    return [("class:prompt", prompt)]


def make_italic(str_):
    """Use Rich's rich text to make string italic."""
    return f"[italic]{str_}[/italic]"


def options_as_help(options: list[str], depth: int = 1) -> str:
    """Return the list of options in a nice format for display.

    Input options might be: ["Action"  , "SomethingElse"  , "Done"]
    first, we convert to    ["(A)ction", "(S)omethingElse", "(D)one"]
    then we render to      "[italic](A)nAction[/italic], "

    Similarly, when called with: ["title", "tag", "back"] AND depth of *2*:
    we convert to                ["(ti)tle", "(ta)g", "(b)ack"]
    and then render to          "[italic](ti)tle[/italic] [italic](ta)g[/italic]..."
    """
    initial_letters = [f"({option[0:depth]}){option[depth:]}" for option in options]
    return ", ".join([make_italic(option) for option in initial_letters])


def goodbye(console) -> None:
    """Print a nice thank you message on successful exit from any command/event-loop."""
    console.print(
        "[italic]Thanks, Gracias, Merci, Danka, ありがとう, спасибо, Köszönöm...![/]\n",
    )
    sys.exit(0)



"""All abstract data types to interact with Raindrop's API."""
from __future__ import annotations

import datetime
import enum
import json
from pathlib import Path
from typing import Any, ClassVar, Optional, Sequence

from dateutil.parser import parse as dateparse
from jashin.dictattr import DictModel, ItemAttr, SequenceAttr

from .api import API

__all__ = [
    "Access",
    "AccessLevel",
    "BrokenLevel",
    "Collection",
    "CollectionRef",
    "DictModel",
    "FontColor",
    "Group",
    "Raindrop",
    "RaindropType",
    "Tag",
    "User",
    "UserConfig",
    "UserFiles",
    "UserRef",
    "View",
]

# Base URL
URL = "https://api.raindrop.io/rest/v1/{path}"


class AccessLevel(enum.IntEnum):
    """Map the Access levels defined by Raindrop's API."""

    readonly = 1
    collaborator_read = 2
    collaborator_write = 3
    owner = 4


class CacheStatus(enum.Enum):
    """Represents the various states the cache of a Raindrop might be in."""

    ready = "ready"
    retry = "retry"
    failed = "failed"
    invalid_origin = "invalid-origin"
    invalid_timeout = "invalid-timeout"
    invalid_size = "invalid-size"


class View(enum.Enum):
    """Map the names of the views for Raindrop's API."""

    list = "list"
    simple = "simple"
    grid = "grid"
    masonry = "masonry"


class CollectionRef(DictModel):
    """Abstract data type for a Raindrop Collection reference."""

    All: ClassVar[CollectionRef]
    Trash: ClassVar[CollectionRef]
    Unsorted: ClassVar[CollectionRef]

    id = ItemAttr[int](name="$id")


# We define the 3 "system" collections in the Raindrop environment:
CollectionRef.All = CollectionRef({"$id": 0})
CollectionRef.Trash = CollectionRef({"$id": -99})
CollectionRef.Unsorted = CollectionRef({"$id": -1})


class UserRef(DictModel):
    """Represents reference to :class:`User` object."""

    id = ItemAttr[int](name="$id")


class Access(DictModel):
    """Represents Access control of Collections."""

    level = ItemAttr(AccessLevel)
    draggable = ItemAttr[bool]()


class Collection(DictModel):
    """Represents a concrete Raindrop Collection."""

    id = ItemAttr[int](name="_id")
    access = ItemAttr(Access)
    collaborators = ItemAttr[Optional[list[Any]]](default=None)
    color = ItemAttr[Optional[str]](default=None)
    count = ItemAttr[int]()
    cover = ItemAttr[list[str]]()
    created = ItemAttr(dateparse)
    expanded = ItemAttr[bool]()
    lastUpdate = ItemAttr(dateparse)
    parent = ItemAttr[Optional[CollectionRef]](CollectionRef, default=None)
    public = ItemAttr[bool]()
    sort = ItemAttr[int]()
    title = ItemAttr[str]()
    user = ItemAttr(UserRef)
    view = ItemAttr(View)

    @classmethod
    def get_roots(cls, api: API) -> Sequence[Collection]:
        """Get "root" Raindrop collections."""
        ret = api.get(URL.format(path="collections"))
        items = ret.json()["items"]
        return [cls(item) for item in items]

    @classmethod
    def get_childrens(cls, api: API) -> Sequence[Collection]:
        """Get the "child" Raindrop collections (ie. all below root level)."""
        ret = api.get(URL.format(path="collections/childrens"))
        items = ret.json()["items"]
        return [cls(item) for item in items]

    @classmethod
    def get_collections(cls, api: API) -> Sequence[Collection]:
        """Query *ALL* collections.

        (hiding the distinction between root/child collections)
        """
        return cls.get_roots(api) + cls.get_childrens(api)

    @classmethod
    def get(cls, api: API, id: int) -> Collection:
        """Return a Raindrop collection based on it's id."""
        url = URL.format(path=f"collection/{id}")
        item = api.get(url).json()["item"]
        return cls(item)

    @classmethod
    def create(
        cls,
        api: API,
        view: Optional[View] = None,
        title: Optional[str] = None,
        sort: Optional[int] = None,
        public: Optional[bool] = None,
        parent: Optional[int] = None,
        cover: Optional[Sequence[str]] = None,
    ) -> Collection:
        """Create a new Raindrop collection."""
        args: dict[str, Any] = {}
        if view is not None:
            args["view"] = view
        if title is not None:
            args["title"] = title
        if sort is not None:
            args["sort"] = sort
        if public is not None:
            args["public"] = public
        if parent is not None:
            args["parent"] = parent
        if cover is not None:
            args["cover"] = cover

        url = URL.format(path="collection")
        item = api.post(url, json=args).json()["item"]
        return Collection(item)

    @classmethod
    def update(
        cls,
        api: API,
        id: int,
        expanded: Optional[bool] = None,
        view: Optional[View] = None,
        title: Optional[str] = None,
        sort: Optional[int] = None,
        public: Optional[bool] = None,
        parent: Optional[int] = None,
        cover: Optional[Sequence[str]] = None,
    ) -> Collection:
        """Update an existing Raindrop collection with any of the attribute values provided."""
        args: dict[str, Any] = {}
        for attr in ["expanded", "view", "title", "sort", "public", "parent", "cover"]:
            if (value := locals().get(attr)) is not None:
                args[attr] = value
        url = URL.format(path=f"collection/{id}")
        item = api.put(url, json=args).json()["item"]
        return Collection(item)

    @classmethod
    def remove(cls, api: API, id: int) -> None:
        """Remove/delete a Raindrop collection."""
        api.delete(URL.format(path=f"collection/{id}"), json={})

    @classmethod
    def get_or_create(cls, api: API, title: str) -> Collection:
        """Get a Raindrop collection based on it's *title*, creating it doesn't exist.

        Return the ID associated with the collection with specified
        collection title (this doesn't seem to be a supported method
        of the Raindrop API directly). If collection is not found, add
        it!.
        """
        # Note: Since
        for collection in Collection.get_roots(api):
            if title.casefold() == collection.values.get("title").casefold():
                return collection

        # Doesn't exist, create it!
        return Collection.create(api, title=title)


class BrokenLevel(enum.Enum):
    """Enumerate user levels."""

    basic = "basic"
    default = "default"
    strict = "strict"
    off = "off"


class FontColor(enum.Enum):
    """Enumerate user display themes available."""

    sunset = "sunset"
    night = "night"


class UserConfig(DictModel):
    """Abstract data type defining a Raindrop user's configuration."""

    broken_level = ItemAttr(BrokenLevel)
    font_color = ItemAttr[Optional[FontColor]](FontColor, default=None)
    font_size = ItemAttr[int]()
    last_collection = ItemAttr[int]()
    raindrops_view = ItemAttr(View)


class Group(DictModel):
    """Abstract data type defining a Raindrop user group."""

    title = ItemAttr[str]()
    hidden = ItemAttr[bool]()
    sort = ItemAttr[int]()
    collectionids = SequenceAttr[int](name="collections")


class UserFiles(DictModel):
    """Abstract data type defining a file associated with a user (?)."""

    used = ItemAttr[int]()
    size = ItemAttr[int]()
    lastCheckPoint = ItemAttr(dateparse)


class User(DictModel):
    """User."""

    id = ItemAttr[int](name="_id")
    config = ItemAttr(UserConfig)
    email = ItemAttr[str]()
    email_MD5 = ItemAttr[str]()
    files = ItemAttr(UserFiles)
    fullName = ItemAttr[str]()
    groups = SequenceAttr(Group)
    password = ItemAttr[bool]()
    pro = ItemAttr[bool]()
    registered = ItemAttr(dateparse)

    @classmethod
    def get(cls, api: API) -> User:
        """Get all the information about a specific Raindrop user."""
        user = api.get(URL.format(path="user")).json()["user"]
        return cls(user)


class SystemCollection(DictModel):
    """Represents a Raindrop System Collection, eg. Unsorted, Trash etc."""

    id = ItemAttr[int](name="_id")
    count = ItemAttr[int]()
    title = ItemAttr[str]()

    # The Raindrop API doesn't give us the names/titles of the
    # "system" collections, let's provide them here.
    CollectionRefsTitles = {
        CollectionRef.All.id: "All",
        CollectionRef.Trash.id: "Trash",
        CollectionRef.Unsorted.id: "Unsorted",
    }

    @classmethod
    def get_status(cls, api: API) -> User:
        """Get the title and counts of Raindrops across all 3 "system" collections."""

        def _add_title(item: dict) -> dict:
            """Add the title (since we don't get from the API)."""
            item["title"] = cls.CollectionRefsTitles[item["_id"]]
            return item

        items = api.get(URL.format(path="user/stats")).json()["items"]
        return [cls(_add_title(item)) for item in items]


class RaindropType(enum.Enum):
    """Map the types of Raindrop bookmarks possible (ie. what type of content they hold)."""

    link = "link"
    article = "article"
    image = "image"
    video = "video"
    document = "document"
    audio = "audio"


class File(DictModel):
    """Represents the attributes associated with a file within a document-based Raindrop."""

    name = ItemAttr[str]()
    size = ItemAttr[int]()  # bytes
    type = ItemAttr[str]()


class Cache(DictModel):
    """Represents the cache information of Raindrop."""

    status = ItemAttr[CacheStatus]()
    size = ItemAttr[int]()  # bytes..
    created = ItemAttr[str]()


class CreatorRef(DictModel):
    """Represents original creator of the Raindrop if different from the current user (ie. shared collections)."""

    id = ItemAttr[int](name="_id")
    fullName = ItemAttr[str]


class Raindrop(DictModel):
    """Core class of a Raindrop bookmark 'item'."""

    # "Main" fields (per https://developer.raindrop.io/v1/raindrops)
    id = ItemAttr[int](name="_id")
    collection = ItemAttr(CollectionRef)
    cover = ItemAttr[str]()
    created = ItemAttr(dateparse)
    domain = ItemAttr[str]()
    excerpt = ItemAttr[str]()  # aka Description on the Raindrop UI.
    lastUpdate = ItemAttr(dateparse)
    link = ItemAttr[str]()
    media = ItemAttr[Sequence[dict[str, Any]]]()
    tags = ItemAttr[Sequence[str]]()
    title = ItemAttr[str]()
    type = ItemAttr(RaindropType)
    user = ItemAttr(UserRef)

    # "Other" fields:
    broken = ItemAttr[bool]()
    cache = ItemAttr[Cache]
    creatorRef = ItemAttr(CreatorRef)
    file = ItemAttr(File)
    important = ItemAttr[bool]()  # aka marked as Favorite.
    # highlights?

    @classmethod
    def get(cls, api: API, id: int) -> Raindrop:
        """Return a Raindrop bookmark based on it's id."""
        item = api.get(URL.format(path=f"{id}")).json()["item"]
        return cls(item)

    @classmethod
    def create_link(
        cls,
        api: API,
        link: str,
        collection: Optional[Collection | CollectionRef, int] = None,
        cover: Optional[str] = None,
        created: Optional[datetime.datetime] = None,
        excerpt: Optional[str] = None,
        html: Optional[str] = None,
        important: Optional[bool] = None,
        lastUpdate: Optional[datetime.datetime] = None,
        media: Optional[Sequence[dict[str, Any]]] = None,
        order: Optional[int] = None,
        pleaseParse: bool = True,
        tags: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
        type: Optional[str] = None,
    ) -> Raindrop:
        """Create a new link-type Raindrop bookmark."""
        # NOTE: We have small code style conflict here between Vulture
        # and Ruff, specifically:

        # Vulture will report all the optional variables above as
        # "unused". This is clearly technically true as their only
        # appearance below is as a string to the arg below from which
        # we take the actual value of the inbound argument from
        # locals(), instead of referring to the argument explicitly.
        #
        # However, converting all of these to a (rather lengthy set)
        # of simple statements like:
        #
        # if <arg> is not None:
        #     args[<arg>] = <arg>
        #
        # keeps Vulture happy but now trips the McCabe complexity
        # metric(!) implemented in our Ruff pre-commit pass due to the
        # number of conditionals in a single method!
        #
        # While, we /could/, convert to kwargs, we lose the documentation
        # and verbosity available with explicit arguments.
        #
        # Thus, for now, we *leave* vulture out of pre-commit and
        # simply run it manually as necessary (through our justfile)

        # Setup the args that will be passed to the underlying
        # Raindrop API, only link is absolutely required, rest are
        # optional!
        args: dict[str, Any] = {"link": link}

        if pleaseParse:
            args["pleaseParse"] = {}

        for attr in [
            "cover",
            "created",
            "excerpt",
            "html",
            "important",
            "lastUpdate",
            "media",
            "order",
            "tags",
            "title",
            "type",
        ]:
            if (value := locals().get(attr)) is not None:
                args[attr] = value

        if collection is not None:
            # <collection> arg could be *either* an actual collection
            # or simply an int collection "id" already, handle either:
            if isinstance(collection, (Collection, CollectionRef)):
                args["collection"] = {"$id": collection.id}
            else:
                args["collection"] = {"$id": collection}

        url = URL.format(path="raindrop")
        item = api.post(url, json=args).json()["item"]
        return cls(item)

    @classmethod
    def create_file(
        cls,
        api: API,
        path: Path,
        content_type: str,
        collection: CollectionRef = CollectionRef.Unsorted,
        tags: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
    ) -> Raindrop:
        """Create a new file-based Raindrop bookmark."""
        url = URL.format(path="raindrop/file")

        # Structure here confirmed through communication with RustemM
        # on 2022-11-29 and his subsequent update to API docs.
        data = {"collectionId": str(collection.id)}
        files = {"file": (path.name, open(path, "rb"), content_type)}
        results = api.put_file(url, path, data, files).json()["item"]
        raindrop = cls(results)

        # The Raindrop API's "Create Raindrop From File" does not
        # allow us to set other attributes, thus, we need to check if
        # any of the possible attributes need to be set and do so
        # explicitly with another call:
        args: dict[str, Any] = {}
        if title is not None:
            args["title"] = title
        if tags is not None:
            args["tags"] = tags
        if args:
            url = URL.format(path=f"raindrop/{raindrop.id}")
            item = api.put(url, json=args).json()["item"]
            return cls(item)
        else:
            return raindrop

    @classmethod
    def update(
        cls,
        api: API,
        id: int,
        collection: Optional[Collection | CollectionRef, int] = None,
        cover: Optional[str] = None,
        created: Optional[datetime.datetime] = None,
        excerpt: Optional[str] = None,
        html: Optional[str] = None,
        important: Optional[bool] = None,
        lastUpdate: Optional[datetime.datetime] = None,
        link: Optional[str] = None,
        media: Optional[Sequence[dict[str, Any]]] = None,
        order: Optional[int] = None,
        pleaseParse: Optional[bool] = False,
        tags: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
        type: Optional[str] = None,
    ) -> Raindrop:
        """Update an existing Raindrop bookmark with any of the attribute values provided."""
        # Setup the args that will be passed to the underlying Raindrop API
        # optional!
        args: dict[str, Any] = {}

        if pleaseParse:
            args["pleaseParse"] = {}

        for attr in [
            "cover",
            "created",
            "excerpt",
            "html",
            "important",
            "lastUpdate",
            "link",
            "media",
            "order",
            "tags",
            "title",
            "type",
        ]:
            if (value := locals().get(attr)) is not None:
                args[attr] = value

        if collection is not None:
            # <collection> arg could be *either* an actual collection
            # or simply an int collection "id" already, handle either:
            if isinstance(collection, (Collection, CollectionRef)):
                args["collection"] = collection.id
            else:
                args["collection"] = collection

        url = URL.format(path=f"raindrop/{id}")
        item = api.put(url, json=args).json()["item"]
        return cls(item)

    @classmethod
    def remove(cls, api: API, id: int) -> None:
        """Remove/delete a Raindrop bookmark."""
        api.delete(URL.format(path=f"raindrop/{id}"), json={})

    @classmethod
    def search(
        cls,
        api: API,
        collection: CollectionRef = CollectionRef.All,
        page: int = 0,
        perpage: int = 50,
        word: Optional[str] = None,
        tag: Optional[str] = None,
        important: Optional[bool] = None,
    ) -> list[Raindrop]:
        """Search for bookmarks in the specified collection with key word, tag or importance parms."""
        args: list[dict[str, Any]] = []
        if word is not None:
            args.append({"key": "word", "val": word})
        if tag is not None:
            args.append({"key": "tag", "val": tag})
        if important is not None:
            args.append({"key": "important", "val": important})

        params = {"search": json.dumps(args), "perpage": perpage, "page": page}
        url = URL.format(path=f"raindrops/{collection.id}")
        results = api.get(url, params=params).json()
        return [cls(item) for item in results["items"]]


class Tag(DictModel):
    """Represents existing Tags, either all or just a specific collection."""

    tag = ItemAttr[str](name="_id")
    count = ItemAttr[int]()

    @classmethod
    def get(cls, api: API, collection_id: int = None) -> list[Tag]:
        """Get all the tags currently defined, either in a specific collections or across all collections."""
        url = URL.format(path="tags")
        if collection_id:
            url += "/" + str(collection_id)
        items = api.get(url).json()["items"]
        return [cls(item) for item in items]

    @classmethod
    def remove(cls, api: API, tags: Sequence[str]) -> None:
        """Remove/delete one or more tags."""
        api.delete(URL.format(path="tags"), json={})



"""Abstract data types to support Raindrop CLI."""
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, TypeVar, Union

from raindropiopy.api import Collection

CreateRequest = TypeVar(
    "CreateRequest",
)  # In py3.11, we'll be able to do 'from typing import Self' instead


@dataclass
class CreateRequest:
    """Encapsulate parameters required to create either a link or file-based Raindrop bookmark."""

    title: str = (
        None  # Bookmark title on Raindrop, eg. "This is a really cool link/doc"
    )
    collection: Union[
        str,
        Collection,
    ] = None  # Name of collection (or real) to store bookmark, eg. "My Documents"
    tags: list[
        str
    ] = None  # Optional list of tags to associate, eg. ["'aTag", "Another Tag"]

    # One of the following needs to be specified:
    url: str = None  # URL of link-based Raindrop to create.
    file_path: Path = (
        None  # Absolute path of file to be pushed, eg. /home/me/Documents/foo.pdf
    )

    def name(self) -> str:
        """Return a user viewable name for request irrespective of type."""
        if self.title:
            return self.title
        if self.url:
            return self.url
        if self.file_path:
            return self.file_path.name
        return "-"

    def print(self, print_method: Callable) -> None:
        """Print the request (using the callable), used to present back to the user for confirmation."""
        if self.url:
            print_method(f"URL           : {self.url}")
        elif self.file_path:
            print_method(f'File to send  : "{self.file_path}"')
        print_method(f"With title    : {self.title}")
        print_method(f"To collection : {self.collection}")
        print_method(f"With tags     : {self.tags}")

    @classmethod
    def factory(cls, entry: dict) -> CreateRequest:
        """Return an new instance of CreateRequest based on an inbound dict, ie. from toml."""
        request: CreateRequest = CreateRequest(
            collection=entry.get("collection"),
            tags=entry.get("tags"),
        )

        # Depending on the population of 2 attributes, we conclude
        # what type of Raindrop we're creating, ie. a standard
        # link/url-based one or a "file"-based one.
        if entry.get("file_path"):
            request.file_path = Path(entry.get("file_path"))

            # Get default title from the file name if we don't have a
            # title provided in the inbound dict:
            request.title = entry.get("title", request.file_path.stem)

        elif entry.get("url"):
            request.url = entry.get("url")
            request.title = entry.get("title")

        return request

    def __str__(self):
        """Render a collection as a string (mostly for display)."""
        return_ = list()
        if self.title:
            return_.append(f"Title      : {self.title}")
        if self.url:
            return_.append(f"URL        : {self.url}")
        if self.file_path:
            return_.append(f"File       : {self.file_path}")
        if self.collection:
            return_.append(f"Collection : {self.collection}")
        if self.tags:
            return_.append(f"Tag(s)     : {self.tags}")
        return "\n".join(return_)



"""Create a new Raindrop bookmark."""
from http.client import HTTPConnection
from pathlib import Path
from time import sleep
from typing import Any, Final, Optional
from urllib.parse import urlparse

from prompt_toolkit.completion import WordCompleter
from tomli import load

from raindropiopy.api import API, Collection, Raindrop
from raindropiopy.cli import (
    CONTENT_TYPES,
    PROMPT_STYLE,
    options_as_help,
    prompt,
)
from raindropiopy.cli.commands import (
    get_confirmation,
    get_from_list,
    get_title,
    get_description,
)
from raindropiopy.cli.models.createRequest import CreateRequest
from raindropiopy.cli.models.eventLoop import EventLoop
from raindropiopy.cli.models.spinner import Spinner


def _create_file(api: API, request: CreateRequest) -> bool:
    """Create a FILE-based Raindrop."""
    args: dict[str, Any] = {}
    if request.title:
        args["title"] = request.title
    if request.tags:
        args["tags"] = request.tags
    if request.description:
        args["excerpt"] = request.description

    Raindrop.create_file(
        api,
        request.file_path,
        content_type=CONTENT_TYPES.get(request.file_path.suffix),
        collection=request.collection,
        **args,
    )
    return True


def _create_link(api: API, request: CreateRequest) -> None:
    """Create a URL/link-based Raindrop."""
    args: dict[str, Any] = {}
    if request.title:
        args["title"] = request.title
    if request.tags:
        args["tags"] = request.tags
    if request.description:
        args["excerpt"] = request.description

    Raindrop.create_link(
        api,
        request.url,
        collection=request.collection,
        **args,
    )


def _read_files(path_: Path) -> list[Path]:
    """Make sure the inbound directory is good and return all PDF's in it."""
    assert path_.exists(), f"Sorry, '{path_}' doesn't exist!"
    return list(path_.glob("*.pdf"))


def __validate_url(url: str) -> Optional[str]:
    """Validate the url provided, returning a message if invalid, None otherwise."""

    def is_url_invalid(url: str) -> Optional[str]:
        try:
            parts = urlparse(url)
            if all([parts.scheme, parts.netloc]):
                if parts.scheme.lower() not in ("http", "https"):
                    return f"Sorry, URL provided {url} isn't valid (bad protocol)"
                return None
        except ValueError:
            ...
        return "Sorry, that URL isn't in a valid format."

    def is_site_invalid(url, timeout=2) -> Optional[str]:
        """Validate that the url provided actually goes to a live site, return message if not."""
        error = Exception("Sorry, unknown error encountered.")
        parser = urlparse(url)
        host = parser.netloc or parser.path.split("/")[0]
        for port in (80, 443):
            connection = HTTPConnection(host=host, port=port, timeout=timeout)
            try:
                connection.request("HEAD", "/")
                return None
            except Exception as e:
                error = e
            finally:
                connection.close()
        return f"Sorry, that URL isn't running now or doesn't exist, can you get to it in a browser? {error}"

    with Spinner("Validating URL..."):
        if msg := is_url_invalid(url):
            return msg
        if msg := is_site_invalid(url):
            return msg
        return None


def __get_url(el: EventLoop) -> Optional[str]:
    ev_prompt = prompt(("create", "url?"))
    while True:
        url = el.session.prompt(ev_prompt, style=PROMPT_STYLE)
        if url == "?" or url == "" or url is None:
            el.console.print(
                "We need a valid URL here, eg. https://www.python.org",
            )
        elif url == "q":
            return None
        else:
            if msg := __validate_url(url):
                el.console.print(msg)
            else:
                return url


def __get_file(el: EventLoop) -> Optional[str]:
    el_prompt = prompt(("create", "bulk", "upload file?"))
    while True:
        file_ = el.session.prompt(el_prompt, style=PROMPT_STYLE)
        if file_ == "?":
            el.console.print(
                "We need a path to a valid, TOML upload file, eg. '/Users/me/Download/upload.toml'",
            )
        elif file_ == "q":
            return None
        else:
            return file_


def __get_from_files(el: EventLoop, options: list[Path]) -> Optional[str]:
    el_prompt = prompt(("create", "url", "file #?"))
    names = [fp_.name for fp_ in options]
    completer: Final = WordCompleter(names)
    for ith, fp_ in enumerate(options):
        el.console.print(f"{ith:2d} {fp_.name}")
    while True:
        response = el.session.prompt(
            el_prompt,
            completer=completer,
            style=PROMPT_STYLE,
            complete_while_typing=True,
            enable_history_search=False,
        )
        if response == "?":
            el.console.print(", ".join(options))
        else:
            break

    return options[int(response)]


def _is_request_valid(el: EventLoop, request: CreateRequest) -> bool:
    """Return True iff the request is valid."""
    # Error Check: Validate the existence of the specified file for
    if request.file_path:
        if not request.file_path.exists():
            el.console.print(
                f"Sorry, no file with name '{request.file_path}' exists.\n",
            )
            return False

        # Validate the file type
        if request.file_path.suffix not in CONTENT_TYPES:
            el.console.print(
                f"Sorry, file type {request.file_path.suffix} isn't yet supported by Raindrop.io.\n",
            )
            return False

    # Error Check: Validate any tags associated with the request
    if request.tags:
        for tag in request.tags:
            if tag.casefold() not in el.state.tags:
                el.console.print(f"Sorry, {tag=} does not currently exist.\n")
                return False

    # Warning only: check for collection existence
    if request.collection:
        if el.state.find_collection(request.collection) is None:
            el.console.print(
                f"Collection '{request.collection}' doesn't exist "
                "in Raindrop, we'll be adding it from scratch.\n",
            )

    return True


def _prompt_for_request(
    el: EventLoop,
    file: bool = False,
    url: bool = False,
    dir_path: Optional[Path] = Path("~/Downloads/Raindrop"),
) -> CreateRequest | None:
    """Prompt for create new Raindrop request (either link or url).

    Returns request or None (if not confirmed).
    """
    request = CreateRequest()

    # Prompts differ whether or not we're creating a file or link-based Raindrop.
    if url:
        request.url = __get_url(el)
        if request.url is None:
            return None
        if not _is_request_valid(el, request):
            return None
    else:
        files = _read_files(dir_path.expanduser())
        request.file_path = __get_from_files(
            el,
            sorted(files, key=lambda fp_: fp_.name),
        )

    # Get a Title:
    request.title = get_title(el, ("create", "title?"))
    if request.title is None:  # User asked to quit the interaction..
        return None

    # Get a Description:
    request.description = get_description(el, ("create", "(description?)"))
    if request.description is None:  # User asked to quit the interaction..
        return None

    # Get a Collection: (these are the same across raindrop types)
    request.collection = get_from_list(
        el,
        ("create", "collection"),
        list(el.state.get_collection_titles(exclude_unsorted=True)),
    )
    if request.collection is None:  # User asked to quit the interaction..
        return None

    # Get a set of Tags to associate with the raindrop:
    request.tags = get_from_list(el, ("create", "tag(s)"), list(el.state.tags))
    if request.tags is None:  # User asked to quit the interaction..
        return None

    # Confirm the values that we just received are good to go...
    request.print(el.console.print)
    if get_confirmation(el, "Is this correct?"):
        return request
    return None  # User declined to go through with the add..


def _add_single(
    el: EventLoop,
    file: bool = False,
    url: bool = False,
    request: CreateRequest = None,
    interstitial: int = 1,
) -> bool:
    """Create either a link or file-based Raindrop, if we don't have a request yet, get one."""
    assert (
        file or url
    ) or request, "Sorry, either a file/url flag or an create request is required"
    if not request:
        request: CreateRequest = _prompt_for_request(el, file=file, url=url)
        if not request:
            return False  # User might have requested that we could still get out..

    # Convert from /name/ of collection user entered to an /instance/
    # of a Collection; if necessary, creating a new one through the
    # respective Raindrop API.
    request.collection = Collection.get_or_create(el.state.api, request.collection)

    # Push it up!
    with Spinner(f"Adding Raindrop -> {request.name()}..."):
        create_method = _create_link if url else _create_file
        create_method(el.state.api, request)

    # Be nice to raindrop.io and wait a bit..
    if interstitial:
        sleep(interstitial)

    return True


def _add_bulk(el: EventLoop) -> None:

    while True:  # ie, Until we get a valid file..
        fn_request = __get_file(el)
        if fn_request is None:
            return None

        fp_request: Path = Path(fn_request).expanduser()
        if not fp_request.exists():
            el.console.print(
                f"Sorry, unable to find request_toml file: '{fp_request}'\n",
            )
            continue

        with open(fp_request, "rb") as fh_request:
            requests: list[CreateRequest] = [
                CreateRequest.factory(entry)
                for entry in load(fh_request).get("requests", [])
            ]

        # Filter down to only valid entries:
        requests: list[CreateRequest] = [
            req for req in requests if _is_request_valid(el, req)
        ]
        if not requests:
            el.console.print("Sorry, no valid requests found.\n")
            break

        # Confirm that we're good to go
        if not get_confirmation(
            el,
            f"Ready to create {len(requests)} valid requests, Ok?",
        ):
            return None

        # Add em!
        return sum([_add_single(el, None, request) for request in requests])


def iteration(el: EventLoop) -> bool:
    """Run a single iteration of our command/event-loop.

    Returns True if we're done, otherwise, keep asking for 'create's.
    """
    options: Final = ("url", "file", "multiple", "back/.")
    options_title: Final = options_as_help(options)
    el.console.print(options_title)
    response = el.session.prompt(
        prompt(("create",)),
        completer=WordCompleter(options),
        style=PROMPT_STYLE,
        complete_while_typing=True,
        enable_history_search=False,
    )

    if response.casefold() in ("back", "b", "."):
        return False

    elif response.casefold() in ("?",):
        el.console.print(options_title)
        return True

    elif response.casefold() in ("url", "u"):
        _add_single(el, url=True)
        return True

    elif response.casefold() in ("file", "f"):
        _add_single(el, file=True)
        return True

    elif response.casefold() in ("multiple", "m"):
        _add_bulk(el)
        return True

    else:
        el.console.print(f"Sorry, must be one of {', '.join(options)}.")


def process(el: EventLoop) -> None:
    """Implement controller log for adding bookmark(s) from the terminal."""
    while iteration(el):
        ...



"""Create a new Raindrop bookmark."""

from prompt_toolkit.completion import WordCompleter
from rich import print

# from raindropiopy.api import Collection, CollectionRef, Raindrop
from raindropiopy.cli import PROMPT_STYLE, prompt
from raindropiopy.cli.models.searchState import SearchState, WILDCARD
from raindropiopy.cli.commands import get_collection_s
from raindropiopy.cli.commands.help import help_search
from raindropiopy.cli.commands.view_edit import process as process_view_edit
from raindropiopy.cli.models.eventLoop import EventLoop
from raindropiopy.cli.models.spinner import Spinner


def __prompt_search_terms(el: EventLoop) -> tuple[bool, str | None]:
    """Prompt for all user response to perform a search, or None if user quits.

    Returns:
    - bool -> True if done with search?
    - str  -> Optional term(s) to search on
    """
    # What tag(s) to do allow for autocomplete?
    search_tags = [f"#{tag}" for tag in el.state.tags]
    completer = WordCompleter(search_tags)
    while True:
        response = el.session.prompt(
            prompt(("search term(s)?",)),
            completer=completer,
            style=PROMPT_STYLE,
            complete_while_typing=True,
            enable_history_search=False,
        )
        if response == "?":
            help_search(el)
        elif response in ("q", "."):
            return True, None
        elif response:
            return False, response

        # Otherwise, we fall through and try again, we need *some* search terms (even if "*" for wildcard)


def _prompt_search(el: EventLoop) -> SearchState | None:
    """Prompt for all responses necessary for a search, ie. terms and collections.

    Returns SearchState for a new search to be performed or None if we're done.
    """
    quit, search_term_s = __prompt_search_terms(el)
    if quit:
        return None

    collection_s = get_collection_s(el, ("search", "collection(s)?"))
    if collection_s == "." or collection_s is None:
        return None

    search_state = (
        SearchState()
    )  # Holds both search request information as we as search results.
    search_state.search = search_term_s
    search_state.collection_s = collection_s.split()
    return search_state


def process(el: EventLoop) -> None:
    """Top-level UI Controller for searching for bookmark(s)."""
    while True:
        search_state = _prompt_search(el)

        if search_state is None:
            return None  # We're REALLY done..

        elif search_state.search == WILDCARD and not search_state.collection_s:
            print("Sorry, wildcard search requires at least one collection.")
            continue

        # Do search and display results (after which we go back for another try.
        collection_text = ", ".join(search_state.collection_s)
        if search_state.search == WILDCARD:
            # Wildcard with at least one collection:
            spinner_text = f"Finding all raindrops in {collection_text}"
        elif not search_state.collection_s:
            # Not a wildcard but don't have a collection specified:
            spinner_text = (
                f"Searching for '{search_state.search}' across all collections"
            )
        else:
            # Not a wildcard but have collection(s) to search over:
            spinner_text = f"Searching for '{search_state.search}' in {collection_text}"

        ################################################################################
        # Do the query and display the results, transferring control over to view/edit
        ################################################################################
        with Spinner(f"{spinner_text}..."):
            search_state.query(el)

        search_state.display_results(el)

        if search_state.results:
            if not process_view_edit(el, search_state):
                return None

        # Otherwise, we go back an try to do another search..



"""Top level API dunder init."""

__all__ = (
    "API",
    "Access",
    "AccessLevel",
    "BrokenLevel",
    "Collection",
    "CollectionRef",
    "DictModel",
    "FontColor",
    "Group",
    "Raindrop",
    "RaindropType",
    "SystemCollection",
    "Tag",
    "User",
    "UserConfig",
    "UserFiles",
    "UserRef",
    "View",
)

from .api import API
from .models import (
    Access,
    AccessLevel,
    BrokenLevel,
    Collection,
    CollectionRef,
    DictModel,
    FontColor,
    Group,
    Raindrop,
    RaindropType,
    SystemCollection,
    Tag,
    User,
    UserConfig,
    UserFiles,
    UserRef,
    View,
)



"""Abstract data types to support Raindrop CLI."""
import os
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, TypeVar

from raindropiopy.api import (
    API,
    Access,
    Collection,
    CollectionRef,
    SystemCollection,
    Tag,
    User,
)
from raindropiopy.cli.models.spinner import Spinner

RaindropState = TypeVar(
    "RaindropState",
)  # In py3.11, we'll be able to do 'from typing import Self' instead


@dataclass
class RaindropState:
    """Encapsulate all aspects for current state of the Raindrop environment."""

    api: API
    created: datetime
    user: User
    collections: list[Collection] = None
    tags: list[str] = None
    refreshed: datetime = None

    def get_collection_titles(
        self,
        exclude_unsorted=False,
        casefold=False,
    ) -> list[str]:
        """Return a sorted list of Collection titles, with or without 'Unsorted' and either casefolded or not."""
        if exclude_unsorted:
            titles = [
                collection.title
                for collection in self.collections
                if collection.id > 0 and collection.title
            ]
        else:
            titles = [
                collection.title for collection in self.collections if collection.title
            ]

        if casefold:
            titles = [title.casefold() for title in titles if title]

        return sorted(titles)

    def find_collection(self, title: str) -> Optional[Collection]:
        """Find the actual Collection object with the title provided."""
        for collection in self.collections:
            if title.casefold() == collection.title.casefold():
                return collection
        return None

    def find_collection_by_id(self, id: int) -> Optional[Collection]:
        """Find the actual Collection object with the *id* provided."""
        for collection in self.collections:
            if collection.id == id:
                return collection
        return None

    def refresh(self) -> bool:
        """Refresh the current state of this Raindrop environment (ie. current collections and tags available)."""

        def __refresh_collections():
            """Gather and set the *COLLECTIONS* associated with the current API user."""
            # This get's us both "root" and "children" collections:
            collections: list[Collection] = Collection.get_collections(self.api)

            # Add the "Unsorted" system collection (we don't care about "Trash" or "All" here)
            for collection in SystemCollection.get_status(self.api):
                if collection.id == CollectionRef.Unsorted.id:
                    unsorted_collection = Collection(
                        {
                            "_id": CollectionRef.Unsorted.id,
                            "count": collection.count,
                            "title": SystemCollection.CollectionRefsTitles[
                                CollectionRef.Unsorted.id
                            ],
                            "access": Access({"level": 4}),
                        },
                    )
                    collections.append(unsorted_collection)

            # Leave our collections in a presentable order, ie. by Title.
            self.collections = sorted(
                collections,
                key=lambda collection: getattr(collection, "title", ""),
            )

        def __refresh_tags():
            """Gather and set the *TAGS* associated with the current API user."""
            # use set to get rid of potential duplicates
            tags: set[str] = set([tag.tag for tag in Tag.get(self.api)])
            self.tags = list(sorted(tags))

        ################################################################################
        with Spinner("Refreshing Raindrop Status..."):
            __refresh_collections()
            __refresh_tags()
            self.refreshed = datetime.utcnow()

        return True

    @classmethod
    def factory(cls) -> RaindropState:
        """Log into Raindrop and return a new, refreshed RaindropState instance."""
        with Spinner("Logging into Raindrop..."):
            api: API = API(
                os.environ["RAINDROP_TOKEN"],
            )  # Setup our connection to Raindrop
            user = User.get(api)  # What user are we currently defined "for"?
        state = RaindropState(api=api, user=user, created=datetime.utcnow())
        state.refresh()
        return state



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['raindropiopy',
 'raindropiopy.api',
 'raindropiopy.cli',
 'raindropiopy.cli.commands',
 'raindropiopy.cli.models']

package_data = \
{'': ['*']}

install_requires = \
['humanize>=4.4.0,<5.0.0',
 'jashin>=0.0.7,<0.0.8',
 'prompt-toolkit>=3.0.36,<4.0.0',
 'pyfiglet>=0.8.post1,<0.9',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'requests-oauthlib>=1.3.1,<2.0.0',
 'rich>=13.2.0,<14.0.0',
 'tomli>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['raindrop-io-py = raindropiopy.cli.cli:main']}

setup_kwargs = {
    'name': 'raindrop-io-py',
    'version': '0.0.10',
    'description': 'API and terminal-based CLI for Raindrop.io bookmark manager',
    'long_description': '#+begin_export html\n<div align="center">\n  <a href="https://choosealicense.com/licenses/mit/">\n    <img alt="MIT License"\n         src="https://img.shields.io/badge/License-MIT-green.svg" />\n  </a>\n\n  <a href="https://www.python.org/">\n    <img alt="Python Version"\n         src="https://img.shields.io/badge/python-3.10+-green" />\n  </a>\n\n  <a href="https://github.com/pre-commit/pre-commit">\n    <img alt="pre-commit"\n       src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit" />\n  </a>\n</div>\n#+end_export\n\n* Raindrop-IO-py\n  Python wrapper for the API to the [[https://raindrop.io][Raindrop.io]] Bookmark Manager as well as a simple command-line interface to perform common operations.\n** Background\n   I wanted to use an existing API for the Raindrop Bookmark Manager ([[https://github.com/atsuoishimoto/python-raindropio][python-raindropio]]) to perform some bulk operations through a simple command-line interface. However, the API available was incomplete and didn\'t contain any interface. Thus, this is a _fork_ and significant extension of [[https://github.com/atsuoishimoto/python-raindropio][python-raindropio]] (thanks [[https://github.com/atsuoishimoto][Atsuo Ishimoto]]!).\n\n   This package includes:\n   - An API providing access to the Raindrop environment (for instance, create, update, delete link/file-based Raindrops; create, update delete Raindrop collections, tags etc.\n   - A terminal-based user-interface that both tests the API as well as providing (me) a fast, simple interface to my Raindrop collections.\n** Status\n   As the API layer is based on a fork of an existing package, it\'s reasonably stable. However, the command-line interface (CLI) is brand new (and lacking tests, ie. "works for me!" ;-).\n** Requirements\n   Requires Python 3.10 or later (well, at least I\'m developing against 3.10.9).\n** Install\n#+begin_src shell\n [.venv] python -m pip install raindrop-io-py\n#+end_src\n** Setup\n    To use this package, besides your own account on [[https://raindrop.io][Raindrop]], you\'ll need to create an ~integration app~ on the Raindrop.io site from which you can create API token(s).\n\n    - Go to [[https://app.raindrop.io/settings/integrations][https://app.draindrop.api/settings/integrations]] and select ~+ create new app~.\n\n    - Give it a descriptive name and then select the app you just created.\n\n    - Select ~Create test token~ and copy the token provided. Note that the basis for calling it a "test" token is that it only gives you access to bookmarks within _your own account_. Raindrop allows you to use their API against other people\'s environments using oAuth (see untested/unsupported flask_oauth file in /examples)\n\n    - Save your token into your environment (we use python-dotenv so a simple .env/.envrc file containing your token should suffice), for example:\n\n#+begin_src shell\n   # If you use direnv or it\'s equivalent, place something like this in a .env file:\n   RAINDROP_TOKEN=01234567890-abcdefghf-aSample-API-Token-01234567890-abcdefghf\n\n   # Or for bash:\n   export RAINDROP_TOKEN=01234567890-abcdefghf-aSample-API-Token-01234567890-abcdefghf\n\n   # Or for fish:\n   set -gx RAINDROP_TOKEN 01234567890-abcdefghf-aSample-API-Token-01234567890-abcdefghf\n\n   # etc...\n#+end_src\n** Examples\n   A full suite of examples are provided in the `examples` directory. Each can be run independently as:\n#+begin_src shell\n   [.venv] % python examples/list_collections.py\n#+end_src\n   or a wrapper script is available to run all of them, in logical order with a small wait to be nice to Raindrop\'s API:\n#+begin_src shell\n   [.venv] % python examples/RUN_ALL.py\n#+end_src\n** API Usage\n   Here are a few examples of API usage (all of these should be able to be executed "as-is"):\n*** Create a New Raindrop Bookmark to a URL\n#+begin_src python\n  import os\n  import sys\n\n  from dotenv import load_dotenv\n\n  from raindropiopy.api import API, Raindrop\n\n  load_dotenv()\n\n  with API(os.environ["RAINDROP_TOKEN"]) as api:\n      link, title = "https://www.python.org/", "Our Benevolent Dictator\'s Creation"\n      print(f"Creating Raindrop to: \'{link}\' with title: \'{title}\'...", flush=True, end="")\n      raindrop = Raindrop.create_link(api, link=link, title=title, tags=["abc", "def"])\n      print(f"Done, id={raindrop.id}")\n\n#+end_src\n    (after this has executed, go to your Raindrop.io environment (site or app) and you should see this Raindrop to python.org available)\n*** Create a New Raindrop Collection\n#+begin_src python\n  import os\n  import sys\n  from datetime import datetime\n  from getpass import getuser\n\n  from dotenv import load_dotenv\n\n  from raindropiopy.api import API, Collection\n\n  load_dotenv()\n\n  with API(os.environ["RAINDROP_TOKEN"]) as api:\n      title = f"TEST Collection ({getuser()}@{datetime.now():%Y-%m-%dT%H:%M:%S})"\n      print(f"Creating collection: \'{title}\'...", flush=True, end="")\n      collection = Collection.create(api, title=title)\n      print(f"Done, {collection.id=}.")\n#+end_src\n    (after this has executed, go to your Raindrop.io environment (site or app) and you should see this collection available)\n*** Display All Bookmarks from the *Unsorted* Raindrop Collection\n#+begin_src python\n  import os\n  from dotenv import load_dotenv\n\n  from raindropiopy.api import API, CollectionRef, Raindrop\n\n  load_dotenv()\n\n  with API(os.environ["RAINDROP_TOKEN"]) as api:\n      page = 0\n      while (items := Raindrop.search(api, collection=CollectionRef.Unsorted, page=page)):\n          for item in items:\n              print(item.title)\n          page += 1\n#+end_src\n** Command-Line Interface Usage\n#+begin_src shell\n   [.venv] % raindropiopy\n#+end_src\n Note: remember to setup RAINDROP_TOKEN in your environment!\n** Acknowledgments\n   - [[https://github.com/atsuoishimoto/python-raindropio][python-raindropio]] from [[https://github.com/atsuoishimoto][Atsuo Ishimoto]].\n** License\n   The project is licensed under the MIT License.\n** Release History\n*** Unreleased\n*** v0.0.10 - 2023-02-05\n    - ADDED: Ability to specify raindrop field: Description on a created Raindrop (either file or link-based).\n    - ADDED: Ability to re-query existing search results (eg. after changes) and smoothed out post-search interactions.\n*** v0.0.9 - 2023-02-04\n    - ADDED: An ability to view, edit and delete raindrops returned from a search.\n    - ADDED: A simple RUN_ALL.py script to the examples directory to...well, run all the examples in order!\n    - CHANGED: The display of raindrops returned from a search to include tags and to only show Collection name if all raindrops are across multiple collections.\n*** v0.0.8 - 2023-01-25\n    - CHANGED: Added simple version method in root package:\n#+begin_src python\n  from raindropiopy import version\n  print(version())\n#+end_src\n*** v0.0.7 - 2023-01-25\n    - CHANGED: Moved from keeping README in markdown to org file format. Incorporated package\'s ChangeLog into README as well (at the bottom).\n    - CHANGED: Added new manage.py release automation capability (internal only, nothing public-facing).\n*** v0.0.6 - 2023-01-22\n    - FIXED: CLI autocomplete now works again after adding support for "single-letter" command-shortcuts.\n    - ADDED: A set of missing attributes to the Raindrop API model type, eg. file, cache etc. Only attribute still missing is `highlights`.\n*** v0.0.5 - 2023-01-21\n    - ADDED: Support use of [[https://github.com/jendrikseipp/vulture][Vulture]] for dead-code analysis (not in pre-commit through due to conflict with ruff\'s McCabe complexity metric)\n    - CHANGED: Moved internal module name to match that of package name. Since we couldn\'t use raindroppy as a package name on PyPI due to similarities with existing packages (one of which was for a *crypto* package), we renamed this package to raindrop-io-py. In concert, the internal module is now ~raindropiopy~:\n#+begin_src python\n   from raindroiopy.api import API\n#+end_src\n    - FIXED: Sample file upload specification in examples/create_raindrop_file.py is now correct.\n',
    'author': 'Peter Borocz',
    'author_email': 'peter.borocz+raindrop-io-py@google.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/PBorocz/raindrop-io-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Add/Create a new file-based bookmark to Raindrop."""
from datetime import datetime
from typing import Final

from humanize import naturaltime
from prompt_toolkit.completion import WordCompleter
from rich.table import Table

from raindropiopy.api.models import Collection, CollectionRef
from raindropiopy.cli import (
    COLOR_TABLE_COLUMN_1,
    COLOR_TABLE_COLUMN_2,
    PROMPT_STYLE,
    options_as_help,
    prompt,
)
from raindropiopy.cli.models.eventLoop import EventLoop


def get_total_raindrops(collections: list[Collection, CollectionRef]) -> int:
    """Return the total number of Raindrops *associated with named collections*."""
    return sum([collection.count for collection in collections if collection.id > 0])


def _show_status(el: EventLoop) -> None:
    """UI Controller for displaying current status."""
    human_diff = naturaltime(datetime.utcnow() - el.state.refreshed)
    table = Table(title=None, show_header=False)
    table.add_column("parm", style=COLOR_TABLE_COLUMN_1, no_wrap=True)
    table.add_column("data", style=COLOR_TABLE_COLUMN_2, justify="right")
    table.add_row("Active User", f"{el.state.user.fullName}")
    table.add_row("Raindrops", f"{get_total_raindrops(el.state.collections):,d}")
    table.add_row("Collections", f"{len(el.state.collections):,d}")
    table.add_row("Tags", f"{len(el.state.tags):,d}")
    table.add_row("As Of", f"{human_diff}")
    el.console.print(table)


def _show_collections(el: EventLoop) -> None:
    """Displaying current collections."""
    total = get_total_raindrops(el.state.collections)
    table = Table(title=None, show_header=False)
    table.add_column("Collection", style=COLOR_TABLE_COLUMN_1, no_wrap=True)
    table.add_column("Count", style=COLOR_TABLE_COLUMN_2, justify="right")
    for collection in el.state.collections:
        table.add_row(collection.title, f"{collection.count:,d}")
    table.add_section()
    table.add_row("Total Raindrops", f"{total:,d}")
    el.console.print(table)


def _show_tags(el: EventLoop) -> None:
    """Display current tags."""
    total = 0
    table = Table(title=None, show_header=False)
    table.add_column("Tag", style=COLOR_TABLE_COLUMN_1, no_wrap=True)
    for tag in el.state.tags:
        table.add_row(tag)
        total += 1
    table.add_section()
    table.add_row(f"{total:,d}")
    el.console.print(table)


def show_help(el: EventLoop) -> None:
    """Display help about this set of commands."""
    el.console.print("status      : Show current status of Raindrop API connection.")
    el.console.print(
        "collections : Display the Collections currently defined along with count of Raindrops in each.",
    )
    el.console.print("tags        : Display the Tags currently defined.")
    el.console.print(
        "refresh     : Refresh the list of Collections & Tags from Raindrop.",
    )


def process(el: EventLoop) -> None:
    """Controller for managing the Raindrop environment, including showing statistics."""
    while True:
        options: Final = (
            "status",
            "collections",
            "tags",
            "refresh",
            "back/.",
        )
        completer: Final = WordCompleter(options)
        options_title: Final = options_as_help(options)
        while True:
            el.console.print(options_title)
            response = el.session.prompt(
                prompt(("manage",)),
                completer=completer,
                style=PROMPT_STYLE,
                complete_while_typing=True,
                enable_history_search=False,
            )
            # Special cases...
            if response.casefold() in ("back", "b", "."):
                return None
            elif response.casefold() in ("help", "h"):
                show_help(el)

            # Normal cases
            elif response.casefold() in ("?",):
                el.console.print(options_title)
            elif response.casefold() in ("status", "s"):
                _show_status(el)
            elif response.casefold() in ("collections", "c"):
                _show_collections(el)
            elif response.casefold() in ("tags", "t"):
                _show_tags(el)
            elif response.casefold() in ("refresh", "r"):
                el.state.refresh()
            else:
                el.console.print(
                    f"Sorry, must be one of {options_title}",
                )



"""Top level cli.commands dunder init, mostly common methods and data types."""
from dataclasses import dataclass
from typing import Iterable

from rich.table import Table

from raindropiopy.api import API, Collection, CollectionRef, Raindrop
from raindropiopy.cli import COLOR_TABLE_COLUMN_1, COLOR_TABLE_COLUMN_2
from raindropiopy.cli.models.eventLoop import EventLoop

WILDCARD: str = "*"


################################################################################
@dataclass
class SearchState:
    """Encapsulates all items necessary and possible for a Raindrop *search* and associated *results*."""

    # Search REQUEST:
    search: str = None
    collection_s: tuple[str] = None

    # Search RESULTS:
    results: list[Raindrop] = None
    selected: int | None = None

    def get_selected(self, ith: int) -> Raindrop | None:
        """Return the Raindrop associated with the ith entry in our results."""
        self.selected = ith - 1
        try:
            return self.results[self.selected]
        except IndexError:
            return None

    def get(self, ith: int) -> str | None:
        """Return the Raindrop associated with the ith entry in our results."""
        try:
            return self.results[ith]
        except IndexError:
            return None

    def display_results(self, el: EventLoop) -> None:
        """Display the list of raindrops in this package of search results (IF we got any!)."""
        if not self.results:
            if self.search:
                msg = f"Sorry, nothing found for search: '{self.search}'"
                if len(self.collection_s) == 0:
                    msg += " across ALL collections."
                elif len(self.collection_s) == 1:
                    msg += f" in collection: '{self.collection_s[0]}'."
                else:
                    msg += f" across {len(self.collection_s)} collections."
                el.console.print(msg)
            return

        # Do we need to show the "Collections" column?
        collection_ids = set([raindrop.collection.id for _, raindrop in self])
        show_collection_name = True if len(collection_ids) > 1 else False

        # How many tag columns do we have?
        max_tags = max([len(raindrop.tags) for _, raindrop in self])

        # ----------
        # Columns
        # ----------
        table = Table(show_header=False)
        table.add_column("#", justify="center", style=COLOR_TABLE_COLUMN_1)
        if show_collection_name:
            table.add_column(
                "Collection",
                justify="left",
                style=COLOR_TABLE_COLUMN_1,
                no_wrap=True,
            )
        table.add_column("Raindrop", style=COLOR_TABLE_COLUMN_2)

        # Add as many columns as the max number of tags we'll have in our results:
        for ith in range(0, max_tags):
            table.add_column(str(ith), style=COLOR_TABLE_COLUMN_2)

        # ----------
        # Rows..
        # ----------
        for ith, raindrop in self:

            # We always show the the "selector" number:
            row = [f"{ith+1}"]

            # We may or may not show the collection name:
            if show_collection_name:
                collection = el.state.find_collection_by_id(raindrop.collection.id)
                row.append(str(collection.title))

            # We always show the title:
            row.append(str(raindrop.title))

            # Show the tags:
            for tag in sorted(raindrop.tags):
                row.append(tag)

            table.add_row(*row)

        # ----------
        # Finally, render the table
        # ----------
        el.console.print(table)

    def prompt(self) -> str:
        """Determine the appropriate prompt "addendum" based on current search state."""
        if self.selected is not None:
            return f"{self.selected + 1}/{len(self)}"
        return f"-/{len(self)}"

    def query(self, el: EventLoop):
        """Query Raindrop.io across 0, 1 or many collections for the respective search terms."""

        def _query_collection(
            api: API,
            search: str,
            collection: Collection | None,
        ) -> list[Raindrop]:
            # Set the search arguments:
            search_args = {"collection": collection}
            if search != WILDCARD:
                search_args["word"] = search

            results: list[Raindrop] = list()
            page: int = 0
            while raindrops := Raindrop.search(api, page=page, **search_args):
                for raindrop in raindrops:
                    results.append(raindrop)
                page += 1
            return results

        # Query based on whether or not we have specific collection(s) to query over:
        self.results: list[Raindrop] = list()
        if self.collection_s:
            for collection_title in self.collection_s:
                self.results.extend(
                    _query_collection(
                        el.state.api,
                        self.search,
                        el.state.find_collection(collection_title),
                    ),
                )
        else:
            self.results.extend(
                _query_collection(el.state.api, self.search, CollectionRef.All),
            )

    def __iter__(self) -> Iterable[tuple]:
        """Return iterable over entries."""
        return iter(enumerate(self.results))

    def __len__(self) -> int:
        """Return the number of raindrops in our search results."""
        return len(self.results)



"""Top level command-line interface controller."""
import sys
from io import StringIO
from pathlib import Path
from typing import Final

from prompt_toolkit import PromptSession
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.history import FileHistory
from pyfiglet import Figlet
from rich.console import Console

from raindropiopy.cli import (
    PROMPT_STYLE,
    goodbye,
    make_italic,
    options_as_help,
    prompt,
)
from raindropiopy.cli.models.raindropState import RaindropState


# Utility method to return the command-history file path for the user
# (creating directories if necessary)
def _get_user_history_path() -> Path:
    history_path = Path("~/.config/raindropiopy").expanduser()
    history_path.mkdir(parents=True, exist_ok=True)

    history_file = history_path / Path(".cli_history")
    if not history_file.exists():
        open(history_file, "a").close()
    return history_file


class EventLoop:
    """Top-level command-line interface controller/command-loop."""

    def _display_startup_banner(self) -> None:
        banner: str = "Raindrop-io-py"
        welcome: str = (
            f"""{make_italic('<tab>')} to show options/complete | """
            f"""{make_italic('help')} for help | """
            f"""{make_italic('Ctrl-D')}, {make_italic('exit')} or '.' to exit."""
        )
        # We can't use self.console.print here as any the special
        # characters figlet creates will be interpreted by Rich.
        print(Figlet(font="thin").renderText(banner))
        self.console.print(welcome)

    def __init__(self, capture: StringIO = None) -> None:
        """Configure our interface components and display our startup banner."""
        # For testing, allow for an internal capture of Console output through the respective arg.
        self.console = Console(file=capture)
        self.session = PromptSession(history=FileHistory(_get_user_history_path()))
        self.state: None  # Will be populated when we start our event loop.
        self._display_startup_banner()

    def iteration(self):
        """Run a single iteration of our command/event-loop."""
        options: Final = ("search", "create", "manage", "exit/.")
        options_title: Final = options_as_help(options)
        self.console.print(options_title)

        response = self.session.prompt(
            prompt(),
            completer=WordCompleter(options),
            style=PROMPT_STYLE,
            complete_while_typing=True,
            enable_history_search=False,
        )

        if response.casefold() in ("exit", "e", "quit", "q", "."):
            raise KeyboardInterrupt  # Quick way out...

        elif response.casefold() in ("?",):
            # FIXME: This should be a longer help text here
            # (since we print the options at the top of each
            # iteration already)
            self.console.print(options_title)
            return

        # We have a valid command, bring in the right module (yes, statically)
        # and dispatch appropriately.
        process_method = None
        if response.casefold() in ("help", "h"):
            from raindropiopy.cli.commands.help import process as process_method

        elif response.casefold() in ("search", "s"):
            from raindropiopy.cli.commands.search import process as process_method

        elif response.casefold() in ("create", "c"):
            from raindropiopy.cli.commands.create import process as process_method

        elif response.casefold() in ("manage", "m"):
            from raindropiopy.cli.commands.manage import process as process_method

        else:
            # Else case here doesn't matter as if process_method isn't set,
            # we'll simply show the list of commands at the top of the next
            # iteration.
            return

        process_method(self)

    def go(self, state: RaindropState) -> None:
        """Save state and run the top-level menu/event loop prompts."""
        self.state = state

        while True:
            try:
                self.iteration()
            except (KeyboardInterrupt, EOFError):
                goodbye(self.console)
                sys.exit(0)
